import os
import warnings


def parse_bool_from_env(key: str, default: bool):
    """Parse a boolean environment variable"""
    if key not in os.environ:
        return default

    value = os.environ[key].strip().lower()

    if value in ["0", "false", "no", "f", "n"]:
        return False

    if value in ["1", "true", "yes", "t", "y"]:
        return True

    warnings.warn(f"Environment variable {key} expects a boolean value (found {value}) => Defaults to {default}.")
    return default


NUMBA_CACHE = parse_bool_from_env("NUMBA_CACHE", True)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Collection, Iterable, List, Sequence, Union
import warnings

import numpy as np
import tqdm.auto as tqdm

import byotrack  # pylint: disable=cyclic-import
from .parameters import ParametrizedObjectMixin


class PauseableTQDM(tqdm.tqdm):  # pylint: disable=inconsistent-mro
    """Tqdm with a real pause mechanism"""

    def __init__(self, *args, **kwargs):
        super().__init__(self, *args, **kwargs)

        self.last_pause_t = 0.0

    def update(self, n: float | None = 1) -> bool | None:
        if self.last_pause_t != 0.0:
            warnings.warn("Progress bar is paused. You should explicitly call unpause before updating again")
            self.unpause()

        return super().update(n)

    def pause(self, refresh=True):
        if self.disable:
            return

        if self.last_pause_t != 0.0:
            warnings.warn("The progress bar was already paused")
            return

        if refresh:  # By default refresh before doing a pause
            self.refresh()

        self.last_pause_t = self._time()  # type: ignore

    def unpause(self):
        if self.disable:
            return

        if self.last_pause_t == 0.0:
            warnings.warn("The progress bar was not paused")
            return

        delta_t = self._time() - self.last_pause_t  # type: ignore
        self.last_pause_t = 0.0
        if delta_t < 0.0:
            warnings.warn("Found negative pause time, ignoring...")
            return

        self.start_t += delta_t
        self.last_print_t += delta_t


class Tracker(ABC, ParametrizedObjectMixin):  # pylint: disable=too-few-public-methods
    """Base abstract tracker class

    A tracker can be run on a whole video and produces tracks

    """

    @abstractmethod
    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Collection[byotrack.Track]:
        """Run the tracker on a whole video

        Args:
            video (Sequence[np.ndarray] | np.ndarray): Sequence of T frames (array).
                Each array is expected to have a shape ([D, ]H, W, C)

        Returns:
            Collection[byotrack.Track]: Tracks of particles

        """


class MultiStepTracker(Tracker):  # pylint: disable=too-few-public-methods
    """Multi step tracker: split the tracking task into Detect / Link / Refine

    Attributes:
        detector (byotrack.Detector): Performs the detection on the video
        linker (byotrack.Linker): Links detections through time
        refiners (Sequence[byotrack.Refiner]): Optional refinement steps
            Empty by default

    """

    def __init__(
        self, detector: byotrack.Detector, linker: byotrack.Linker, refiners: Iterable[byotrack.Refiner] = ()
    ) -> None:
        super().__init__()

        self.detector = detector
        self.linker = linker
        self.refiners = refiners

    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Collection[byotrack.Track]:
        detections = self.detector.run(video)
        tracks = self.linker.run(video, detections)
        for refiner in self.refiners:
            tracks = refiner.run(video, tracks)

        # Check produced tracks
        byotrack.Track.check_tracks(tracks, warn=True)

        return tracks


class BatchMultiStepTracker(MultiStepTracker):  # pylint: disable=too-few-public-methods
    """Online Tracking with detect/link/refine framework

    It only works with BatchDetector and OnlineLinker: it reduces RAM usage and computations
    by loading frames only by batch. Each batch is processed by the detector followed
    by the linker. Once a batch is processed, a new one is loaded. The detections are never fully stored,
    neither the video.

    Tracks refining is done offline.

    Attributes:
        detector (byotrack.BatchDetector): Performs the detection on the video by batch
            It defines the batch size to use
        linker (byotrack.OnlineLinker): Links detections one frame at time
        refiners (Sequence[byotrack.Refiner]): Optional refinement steps
            Empty by default

    """

    def __init__(
        self, detector: byotrack.BatchDetector, linker: byotrack.OnlineLinker, refiners: Iterable[byotrack.Refiner] = ()
    ) -> None:
        super().__init__(detector, linker, refiners)
        self.detector: byotrack.BatchDetector
        self.linker: byotrack.OnlineLinker

    def run(self, video: Union[Sequence[np.ndarray], np.ndarray]) -> Collection[byotrack.Track]:
        if len(video) == 0:
            return []

        reader = None
        if self.detector.add_true_frames and isinstance(video, byotrack.Video):
            reader = video.reader

        self.linker.reset()

        detections_sequence: Sequence[byotrack.Detections] = []
        frame_ids: List[int] = []

        detect_bar = PauseableTQDM(desc=self.detector.progress_bar_description, total=len(video))
        link_bar = PauseableTQDM(desc=self.linker.progress_bar_description, total=len(video))
        link_bar.pause()  # Starting with detections

        first = video[0]
        batch = np.zeros((self.detector.batch_size, *first.shape), dtype=first.dtype)
        batch[0] = first
        n = 1
        frame_ids.append(reader.tell() if reader else 0)

        for frame in video[1:]:
            if n >= self.detector.batch_size:
                detections_sequence = self.detector.detect(batch)
                detect_bar.update(n)

                # Switch from detection to linking
                detect_bar.pause()
                link_bar.unpause()

                for i in range(n):
                    detections_sequence[i].frame_id = frame_ids[i]
                    self.linker.update(batch[i], detections_sequence[i])
                    link_bar.update()

                # Switch from linking to detection
                link_bar.pause()
                detect_bar.unpause()

                frame_ids = []
                n = 0

            batch[n] = frame
            frame_ids.append(reader.tell() if reader else len(frame_ids))
            n += 1

        detections_sequence = self.detector.detect(batch[:n])
        detect_bar.update(n)
        detect_bar.close()

        link_bar.unpause()

        for i in range(n):
            detections_sequence[i].frame_id = frame_ids[i]
            self.linker.update(batch[i], detections_sequence[i])
            link_bar.update()

        link_bar.close()

        tracks = self.linker.collect()
        for refiner in self.refiners:
            tracks = refiner.run(video, tracks)

        # Check produced tracks
        byotrack.Track.check_tracks(tracks, warn=True)

        return tracks



"""ByoTrack

Unified python API for biological particle tracking (2D/3D).

Many bioimage informatics tools already implement their own tracking tools (Icy [1], ImageJ [6], ...)
but most of them are implemented in Java which makes it difficult for non-Java developers to experiment
with the code. It is also difficult to integrate deep learning algorithms (mainly developed in Python)
into these software.

We provide a unified python API for tracking that can be easily extended with new (and old) algorithms.
We also provide implementations of well-known algorithms following our API. ByoTrack is based on numpy,
pytorch and numba allowing fast computations with the access to the full python ecosystem.

Overview:
* Video
    * Able to read most classical format (supported by opencv) + tiff
    * Supports 2D and 3D.
* Particle Tracking
    * MultiStepTracker (Detect / Link / Refine)
* Particle Detections
    * Wavelet Detector [2] (Similar as the one in Icy [1] but coded in pytorch)
    * Stardist [3] (Inference only. Training should be done with the
        [official implementation](https://github.com/stardist/stardist))
* Particle Linking
    * Nearest neighbors using optical flow, kalman filters or both (KOFT) [9]
    * EMHT [4] (Wraps the implementation in Icy [1], requires Icy to be installed)
    * u-track / TrackMate [7] (Wraps the TrackMate [6, 8] implementation in ImageJ/Fiji, requires Fiji to be installed)
* Tracks Refining
    * Cleaning
    * EMC2 [5]: Track stitching (gap closing)
    * Interpolate missing positions
* Optical Flow
    * Support for Open-CV and Scikit-Image algorithms. Can be used for particle linking, track stitching
    and interpolations.
* Datasets
    * Supports for loading annotations/video from datasets.
    * Cell Tracking Challenge (CTC) [10]
    * SINETRA [11]
* Metrics:
    * Support for some segmentation/detection/tracking metrics.
        Currently, only CTC metrics are provided. More to come...


Getting started:
----------------

```python
import byotrack

# Load some specific implementations
from byotrack.implementation.detector.wavelet import WaveletDetector
from byotrack.implementation.linker.icy_emht import IcyEMHTLinker
from byotrack.implementation.refiner.cleaner import Cleaner
from byotrack.implementation.refiner.stitching import EMC2Stitcher

# Read a video from a path, normalize and aggregate channels
video = byotrack.Video(video_path)
transform_config = VideoTransformConfig(aggregate=True, normalize=True, q_min=0.01, q_max=0.999)
video.set_transform(transform_config)

# Create a multi step tracker
## First the detector
## Smaller scale <=> search for smaller spots
## The noise threshold is linear with k. If you increase it, you will retrieve less spots.
detector = WaveletDetector(scale=1, k=3.0, min_area=5)

## Second the linker
## Hyperparameters are automatically chosen by Icy
linker = IcyEMHTLinker(icy_path)

## Finally refiners
## If needed you can add Cleaning and Stitching operations
refiners = []
if True:
    refiners.append(Cleaner(5, 3.5))  # Split tracks on position jumps and drop small ones
    refiners.append(EMC2Stitcher())  # Merge tracks if they track the same particle

tracker = byotrack.MultiStepTracker(detector, linker, refiners)

# Run the tracker
tracks = tracker.run(video)

# Save tracks
byotrack.Track.save(tracks, output_path)
```

Please refer to the ![official documentation](https://byotrack.readthedocs.io/en/latest/?badge=latest)

"""

# Import API and main features but no implementation
from byotrack._env import NUMBA_CACHE
from byotrack.api.detector.detections import Detections
from byotrack.api.detector.detector import BatchDetector, Detector
from byotrack.api.linker import Linker, OnlineLinker
from byotrack.api.features_extractor import FeaturesExtractor, MultiFeaturesExtractor
from byotrack.api.optical_flow.optical_flow import OpticalFlow
from byotrack.api.refiner import Refiner
from byotrack.api.tracker import BatchMultiStepTracker, MultiStepTracker, Tracker
from byotrack.api.tracks import Track
from byotrack.video import Video, VideoTransformConfig


__version__ = "1.3.11"



from __future__ import annotations

import dataclasses
import os
from typing import overload, Optional, Sequence, Tuple, Union

import numpy as np

from .reader import VideoReader, slice_length
from .transforms import ChannelSelect, ChannelAvg, ScaleAndNormalize


@dataclasses.dataclass
class VideoTransformConfig:
    """Configuration for video transformations

    Attributes:
        aggregate (bool): Aggregate channels
        normalize (bool): Scale and Normalize the video in [0, 1]
        selected_channel (Optional[int]): Channel to use for aggregation
            If None, channel average is done. If any, it performs channel selection
        q_min (float): Minimum quantile to use when scaling the video
        q_max (float): Maximum quantile to use when scaling the video
        smooth_clip (float): Smoothness of the clipping process (log clipping)
            See `ScaleAndNormalize`: it logs clip the highest values on q_max.
            If 0.0, hard clipping is done.
        compute_stats_on (int): Number of frames to use to compute the quantiles.

    """

    aggregate: bool = False
    normalize: bool = False
    selected_channel: Optional[int] = None
    q_min: float = 0.0
    q_max: float = 1.0
    smooth_clip: float = 0.0
    compute_stats_on: int = 50


class Video(Sequence[np.ndarray]):
    """Video: Iterable, indexable and sliceable sequence of frames wrapping a VideoReader

    It wraps VideoReader in order to add video transformation (Channel Aggregation, Scaling, Normalization)
    and to add useful pythonic protocols (Sliceable, Indexable, Iterable).

    Frames are 2D or 3D with a channel axis. It behaves similarly as a 5D/4D numpy array of shape (T[, D], H, W, C).

    Example:
        .. code-block:: python

            import byotrack

            # Read a video (Usually 2D RGB)
            video = byotrack.Video(video_path)

            # Add a transform that will aggregate channel and normalize in [0, 1] the intensities
            transform_config = byotrack.VideoTransformConfig(aggregate=True, normalize=True, q_min=0.01, q_max=0.999)
            video.set_transform(transform_config)

            # Iterate through the video
            for frame in video:
                pass

            # Temporal slicing
            sliced = video[10:50:3]  # Take one frame every three from frame 10 to frame 50.

            # Spatial slicing
            sliced = video[:, 100:200, 150:250]  # All frames on the roi (100:200 x 150:250)


    Attributes:
        ndim (int): Either 4 (2D) or 5 (3D). (T, H, W, C) in 2D or (T, D, H, W, C) in 3D.
        shape (Tuple[int, ...]): Shape of the video (Time, [Depth, ]Height, Width)
        channels (int): Number of channels
        reader (byotrack.VideoReader): Underlying video reader

    """

    def __init__(self, data_source: Union[str, os.PathLike, VideoReader], **kwargs) -> None:
        """Constructor.


        Args:
            data_source (Union[str, os.PathLike, VideoReader]): Source of the data. If a path is given,
                it will be converted in a VideoReader.
            **kwargs: Additional arguments given to the construction of the video reader.

        """
        super().__init__()

        if isinstance(data_source, (str, os.PathLike)):
            self.reader = VideoReader.open(data_source, **kwargs)
        else:
            self.reader = data_source

        self._slices: Tuple[slice, ...] = tuple(slice(None) for _ in (self.reader.length, *self.reader.shape))
        self._channel_aggregator: Optional[Union[ChannelAvg, ChannelSelect]] = None
        self._normalizer: Optional[ScaleAndNormalize] = None

    @property
    def shape(self) -> Tuple[int, ...]:
        return tuple(
            slice_length(slice_, shape) for slice_, shape in zip(self._slices, (self.reader.length, *self.reader.shape))
        )

    @property
    def ndim(self) -> int:
        return 2 + len(self.reader.shape)

    @property
    def channels(self) -> int:
        return self.reader.channels if self._channel_aggregator is None else 1

    def set_transform(self, transform_config: VideoTransformConfig) -> None:
        """Set the transform (channel_selector and normalizer)

        Args:
            transform_config (byotrack.VideoTransformConfig): Configuration of the transformations

        """
        self._channel_aggregator = None
        if transform_config.aggregate:
            if transform_config.selected_channel is not None:
                self._channel_aggregator = ChannelSelect(transform_config.selected_channel)
            else:
                self._channel_aggregator = ChannelAvg()

        self._normalizer = None
        if transform_config.normalize:
            frames = np.asarray(self[: transform_config.compute_stats_on])
            self._normalizer = ScaleAndNormalize(
                transform_config.q_min,
                transform_config.q_max,
                transform_config.smooth_clip,
                transform_config.compute_stats_on,
            )
            self._normalizer.update_stats(frames)

    def transform(self, frame: np.ndarray) -> np.ndarray:
        """Transform a frame using channel aggregation and normalization"""
        if self._channel_aggregator:
            frame = self._channel_aggregator(frame)
        if self._normalizer:
            frame = self._normalizer(frame)

        return frame

    def __len__(self) -> int:
        return slice_length(self._slices[0], self.reader.length)  # length of temporal slice

    @overload
    def __getitem__(self, index: int) -> np.ndarray: ...

    @overload
    def __getitem__(self, slice_: slice) -> Video: ...

    @overload
    def __getitem__(self, slices: Tuple[slice, ...]) -> Video: ...

    def __getitem__(self, key):  # pylint: disable=too-many-branches
        """Indexing and slicing operations

        When indexed, it returns the ith frame in the slice
        When sliced, it duplicates the video (wrapper) with the right slice

        Args:
            key (int | slice | Tuple[slice, ...]): index or slice of the video

        Returns:
            np.ndarray | Video: Frame at index or a shallow copy of the video with the right slice

        """
        if isinstance(key, int):
            start, _, step = self._slices[0].indices(self.reader.length)

            if key < 0:
                key += len(self)

            if key < 0 or key >= len(self):
                raise IndexError(f"Index {key} out of range")

            frame_id = start + key * step

            if frame_id == self.reader.frame_id:
                pass  # Skip expensive seek
            if frame_id == self.reader.frame_id + 1:
                if not self.reader.grab():  # Much faster for cv2: Allows a fast frame by frame reading
                    raise RuntimeError(f"Unable to grab frame {frame_id}")
            else:
                try:
                    self.reader.seek(frame_id)
                except EOFError:
                    raise RuntimeError(f"Unable to seek frame {frame_id}") from None

            return self.transform(self.reader.retrieve()[self._slices[1:]])

        if isinstance(key, slice):
            key = (key,)

        if isinstance(key, tuple):
            if len(key) > len(self._slices):
                raise IndexError("Too many indices for video. Only support 3 dimensions slicing (time, height, width).")

            slices = list(self._slices)
            shapes = (self.reader.length, *self.reader.shape)
            for i, slice_ in enumerate(key):
                if not isinstance(slice_, slice):
                    raise TypeError("Unsupported index for Video. Supports only int, slice and Tuple[slice, ...]")

                slices[i] = compose_slice(self._slices[i], slice_, shapes[i])

            # Duplicate
            other = Video(self.reader)
            other._channel_aggregator = self._channel_aggregator
            other._normalizer = self._normalizer
            other._slices = tuple(slices)

            return other

        raise TypeError("Unsupported index for Video. Supports only int, slice and Tuple[slice, ...]")


def compose_slice(slice_1: slice, slice_2: slice, length: int) -> slice:
    """Compose two slices in the given order"""
    # Compute start, stop, step and length of the first slice
    start_1, _, step_1 = slice_1.indices(length)
    length_1 = slice_length(slice_1, length)

    # Same for the second slice
    start_2, stop_2, step_2 = slice_2.indices(length_1)
    length = slice_length(slice_2, length_1)

    # Compute the final start, stop and step
    start = start_1 + start_2 * step_1
    stop = start_1 + stop_2 * step_1
    step = step_1 * step_2

    if start > stop:
        if stop < 0:  # (5, -1, -1) is empty, instead use (5, None, -1)
            return slice(start, None, step)

    return slice(start, stop, step)



# pylint: disable=duplicate-code

import dataclasses
from typing import Optional, Tuple, Union
import warnings

import numpy as np
import torch
import torch_kf
import torch_kf.ckf

import byotrack

from .base import AssociationMethod, OnlineFlowExtractor
from .kalman_linker import Cost, KalmanLinker, KalmanLinkerParameters, TrackBuilding


@dataclasses.dataclass
class KOFTLinkerParameters(KalmanLinkerParameters):
    """Parameters of KOFTLinker

    Note:
        The merging and splitting features is still experimental.

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
        detection_std (Union[float, torch.Tensor]): Expected measurement noise on the detection process.
            The detection process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension). See `torch_kf.ckf.constant_kalman_filter`.
            Default: 3.0 pixels
        flow_std (Union[float, torch.Tensor]): Expected measurement noise on the optical flow process.
            The optical flow process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension).
            Default: 1.0 pixels
        process_std (Union[float, torch.Tensor]): Expected process noise. See `torch_kf.ckf.constant_kalman_filter`, the
            process is modeled as constant order-th derivative motion. This quantify how much the supposely "constant"
            order-th derivative can change between two consecutive frames. A common rule of thumb is to use
            3 * process_std ~= max_t(| dx^(order)(t+1) - dx^(order)(t)|). It can be provided for each dimension).
            Default: 1.5 pixels
        kalman_order (int): Order of the Kalman filter to use. 0 is for brownian motion (it predicts a 0 velocity)
            1 for directed brownian motion, 2 for accelerated brownian motions, etc...
            Default: 1
        n_valid (int): Number associated detections required to validate the track after its creation.
            Default: 3
        n_gap (int): Number of consecutive frames without association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, [SPARSE_]OPT_HARD, [SPARSE_]OPT_SMOOTH)
            Default: OPT_SMOOTH
        anisotropy (Tuple[float, float, float]): Anisotropy of images (Ratio of the pixel sizes
            for each axis, depth first). This will be used to scale distances. It will only impact
            EUCLIDEAN[_SQ] costs. For probabilistic cost, anisotropy should be already integrated
            in the stds of the kalman filter (providing one std for each dimension).
            Default: (1., 1., 1.)
        cost_method (CostMethod): The cost method to use. It can be provided as a string.
            See `CostMethod`. It also indicates what is the correct unit of `association_threshold`.
            Default: EUCLIDEAN
        track_building (TrackBuilding): Tells the linker how to build the final tracks.
            Either from detections, or from filtered/smoothed positions computed by the
            Kalman filter. See `TrackBuilding`. It can be provided as a string.
            Default: FILTERED
        split_factor (float): Allow splitting of tracks, using a second association step.
            The association threshold in this case is `split_factor * association_threshold`.
            Default: 0.0 (No splits)
        merge_factor (float): Allow merging of tracks, using a second association step.
            The association threshold in this case is `merge_factor * association_threshold`.
            Default: 0.0 (No merges)
        extract_flows_on_detections (bool): If True it extracts the optical flow at the detection location if possible.
            Otherwise it extract the flow from the curent estimate of the track position.
            Default: False
        always_measure_velocity (bool): Update velocity for all tracks even non-linked ones.
            If set to False, it implements KOFT-- from the paper. This is sub-optimal, you should keep it True.
            Default: True
        online_process_std (float): Recomputes the process std online following "A. Genovesio, et al, 2004, October.
            Adaptive gating in Gaussian Bayesian multi-target tracking. ICIP'04. (Vol. 1, pp. 147-150). IEEE."
            Each track has its own process std depending on the errors made in the past. It automatically adjusts to
            process errors, allowing to increase the validation gate. Should be used in conjonction with MAHALANOBIS
            or LIKELIHOOD `cost_method`. As this may be detrimental, it is disabed by default.
            Default: 0.0 (Process_std is constant)
        initial_std_factor (float): The uncertainties on initial velocities/accelerations are set
            to initial_std_factor * process_std. See `KalmanLinker.build_initial_covariance`.
            Having a small factor will prevent handling correctly starting tracks with large initial velocity
            on their first frames.
            Typical values lies between 3.0 to 100.0.
            Default: 10.0

    """

    def __init__(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        association_threshold: float,
        *,
        detection_std: Union[float, torch.Tensor] = 3.0,
        flow_std: Union[float, torch.Tensor] = 1.0,
        process_std: Union[float, torch.Tensor] = 1.5,
        kalman_order: int = 1,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0),
        cost: Union[str, Cost] = Cost.EUCLIDEAN,
        track_building: Union[str, TrackBuilding] = TrackBuilding.FILTERED,
        split_factor: float = 0.0,
        merge_factor: float = 0.0,
        extract_flows_on_detections=False,
        always_measure_velocity=True,
        online_process_std=0.0,
        initial_std_factor=10.0,
    ):
        super().__init__(
            association_threshold=association_threshold,
            detection_std=detection_std,
            process_std=process_std,
            kalman_order=kalman_order,
            n_valid=n_valid,
            n_gap=n_gap,
            association_method=association_method,
            anisotropy=anisotropy,
            cost=cost,
            track_building=track_building,
            split_factor=split_factor,
            merge_factor=merge_factor,
            online_process_std=online_process_std,
            initial_std_factor=initial_std_factor,
        )

        if isinstance(flow_std, float) and min(anisotropy) != max(anisotropy):
            warnings.warn(
                "A single flow_std is provided, but the images are anisotrope. Consider giving one std by dimension."
            )

        self.flow_std = flow_std
        self.extract_flows_on_detections = extract_flows_on_detections
        self.always_measure_velocity = always_measure_velocity

    flow_std: Union[float, torch.Tensor] = 1.0
    extract_flows_on_detections: bool = False
    always_measure_velocity: bool = True


class KOFTLinker(KalmanLinker):
    """Kalman and Optical Flow Tracking [9]

    Motion is modeled with a Kalman filter of a specified order >= 1 (See `torch_kf.ckf`)
    Positions are measured through the detection process. A second update step is performed
    to measure the velocity of all tracks using optical flow.

    Matching is done to optimize the given cost.

    Note:
        This implementation requires torch-kf. (pip install torch-kf)

    See `KalmanLinker` for the other attributes.

    Attributes:
        specs (KOFTLinkerParameters): Parameters specifications of the algorithm.
            See `KOFTLinkerParameters`.
        last_detections (byotrack.Detections): The last detections used in update.
            Optionnaly used to extract flows at the detection positions and not the track state.
            Required for `motion_model`

    """

    progress_bar_description = "KOFT linking"

    def __init__(
        self,
        specs: KOFTLinkerParameters,
        optflow: Optional[byotrack.OpticalFlow] = None,
        features_extractor: Optional[byotrack.FeaturesExtractor] = None,
        save_all=False,
    ) -> None:
        super().__init__(specs, optflow, features_extractor, save_all)

        self.optflow: OnlineFlowExtractor
        self.specs: KOFTLinkerParameters
        assert self.optflow is not None, "KOFT requires an optical flow algorithm"

        self.last_detections = byotrack.Detections(data={"position": torch.empty((0, 2))})

    def reset(self, dim=2) -> None:
        super().reset(dim)

        # Reset the KF initialized by super
        self.kalman_filter = torch_kf.ckf.constant_kalman_filter(
            self.specs.detection_std,
            self.specs.process_std,
            dim=dim,
            order=self.specs.kalman_order + (self.specs.kalman_order == 0),
            approximate=True,  # Approximate so that a flow precisely means the velocity modeled here
        )
        if self.specs.kalman_order == 0:
            # In order 0, we still model velocity, but we always predict it at 0
            self.kalman_filter.process_matrix[dim:, dim:] = 0

        # Doubles the measurement space to measure velocity
        self.kalman_filter.measurement_matrix = torch.eye(dim * 2, self.kalman_filter.state_dim)
        self.kalman_filter.measurement_noise = torch.eye(dim * 2)
        self.kalman_filter.measurement_noise[:dim, :dim] *= self.specs.detection_std**2
        self.kalman_filter.measurement_noise[dim:, dim:] *= self.specs.flow_std**2
        self.kalman_filter = self.kalman_filter.to(self.dtype)

        self.active_states = torch_kf.GaussianState(
            torch.empty((0, self.kalman_filter.state_dim, 1), dtype=self.dtype),
            torch.empty((0, self.kalman_filter.state_dim, self.kalman_filter.state_dim), dtype=self.dtype),
        )
        self.projections = self.kalman_filter.project(self.active_states)

        self.last_detections = byotrack.Detections(data={"position": torch.empty((0, dim))})

    def motion_model(self) -> None:
        if self.optflow.flow_map is None:
            raise RuntimeError("The motion model cannot be applied without a computed flow")

        dim = self.kalman_filter.measure_dim // 2

        # Second update using velocity just before the prediction time
        points = self.active_states.mean[:, :dim, 0].clone()
        if self.specs.extract_flows_on_detections:
            for i, track in enumerate(self.active_tracks):
                if track.detection_ids[-1] != -1:
                    points[i] = self.last_detections.position[track.detection_ids[-1]]

        if not self.specs.always_measure_velocity:  # KOFT--
            mask = torch.full((len(self.active_tracks),), True)
            for i, track in enumerate(self.active_tracks):
                if track.detection_ids[-1] == -1:
                    mask[i] = False

            points = points[mask]

        velocities = torch.tensor(self.optflow.optflow.flow_at(self.optflow.flow_map, points.numpy()))[..., None]

        if not self.specs.always_measure_velocity:  # KOFT--
            self.active_states[mask] = self.kalman_filter.update(
                self.active_states[mask],
                velocities,
                measurement_matrix=self.kalman_filter.measurement_matrix[dim:],
                measurement_noise=self.kalman_filter.measurement_noise[dim:, dim:],
            )
        else:  # KOFT
            self.active_states = self.kalman_filter.update(
                self.active_states,
                velocities,
                measurement_matrix=self.kalman_filter.measurement_matrix[dim:],
                measurement_noise=self.kalman_filter.measurement_noise[dim:, dim:],
            )

        # Replace the registered states after adding optical flow inside of it
        if self.save_all or self.specs.track_building == TrackBuilding.SMOOTHED:
            if self.all_states:
                self.all_states[-1] = self.active_states

        # Use the Kalman filter to predict the current states of each active tracks
        self.active_states = self.kalman_filter.predict(
            self.active_states, process_noise=self.process_noises if self.specs.online_process_std else None
        )

        # Project states for association
        self.projections = self.kalman_filter.project(self.active_states)

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        anisotropy = torch.tensor(self.specs.anisotropy)[-detections.dim :]

        if self.specs.cost == Cost.EUCLIDEAN:
            return (
                torch.cdist(
                    self.projections.mean[:, : detections.dim, 0] * anisotropy, detections.position * anisotropy
                ),
                self.specs.association_threshold,
            )

        if self.specs.cost == Cost.EUCLIDEAN_SQ:
            return (
                torch.cdist(
                    self.projections.mean[:, : detections.dim, 0] * anisotropy, detections.position * anisotropy
                ).pow_(2),
                self.specs.association_threshold**2,
            )

        # Restrict projections onto positions
        # NOTE: Following KOFT initial implem, we use (cov-1)[:2, :2] instead of (cov[:2, :2])-1 for the precision.
        #       It works slightly better, but it needs to be investigated.
        projections = torch_kf.GaussianState(
            self.projections.mean[:, : detections.dim],
            self.projections.covariance[:, : detections.dim, : detections.dim],
            (
                self.projections.precision[:, : detections.dim, : detections.dim]
                if self.projections.precision is not None
                else None
            ),
        )

        if self.specs.cost == Cost.MAHALANOBIS:
            return (
                projections[:, None].mahalanobis(detections.position[None, ..., None]),
                self.specs.association_threshold,
            )
        if self.specs.cost == Cost.MAHALANOBIS_SQ:
            return (
                projections[:, None].mahalanobis_squared(detections.position[None, ..., None]),
                self.specs.association_threshold**2,
            )

        # LIKELIHOOD: cost = -log likelihood
        cost = -projections[:, None].log_likelihood(detections.position[None, ..., None])
        return cost, -torch.log(torch.tensor(self.specs.association_threshold)).item()

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, active_mask: torch.Tensor):
        self.last_detections = detections  # Save detections (May be required)
        positions = detections.position.to(self.dtype)

        # Update the state of associated tracks (unassociated tracks keep the predicted state)
        updated = self.kalman_filter.update(
            self.active_states[self._links[:, 0]],
            positions[self._links[:, 1]][..., None],
            projection=torch_kf.GaussianState(
                self.projections.mean[self._links[:, 0], : detections.dim],
                self.projections.covariance[self._links[:, 0], : detections.dim, : detections.dim],
                None,  # /!\ inv(cov[:2,:2]) != inv(cov)[:2, :2]
            ),
            measurement_matrix=self.kalman_filter.measurement_matrix[: detections.dim],
            measurement_noise=self.kalman_filter.measurement_noise[: detections.dim, : detections.dim],
        )

        if self.specs.online_process_std:
            # XXX: Should be computed after velocity update ?
            # Update the process_noise based on the prediction errors.
            # Err = x_t - Fx_{t-1} (low estimation of errors)
            # Then Q_t = a * (b * Q_t + 1 - b err @ err.T) + (1- a) Q_0
            # In this implem, we fix b = 0.75 (around 4 frames to estimate errors) and a is user-defined.
            errors = self.active_states.mean[self._links[:, 0]] - updated.mean
            self.process_noises[self._links[:, 0]] *= self.specs.online_process_std * 0.75  # Fixed EMA ~4 frames lag
            self.process_noises[self._links[:, 0]] += self.specs.online_process_std * 0.25 * errors @ errors.mT
            self.process_noises[self._links[:, 0]] += (
                1 - self.specs.online_process_std
            ) * self.kalman_filter.process_noise

        self.active_states[self._links[:, 0]] = updated

        # Create new states for unmatched measures
        unmatched_measures = positions[self._unmatched_detections]

        # Build the initial states for tracks:
        # Infinite uncertainty on the prior position => Given positional measurement, we initialize on its position
        #                                               with the positional measurement uncertainty.
        # For derivatives, we assume they are centered on 0 (no bias toward a direction) with an uncertainty of
        # initial_std_factor * process_std:
        # Small factors (~1) are not suited for tracking objects that appears with a large initial velocity.
        # Large factors (>> 1) will create association problem on the second frame, as the predicted
        # position inherit from this uncertainty.
        # As in KOFT, the velocities are measured, it could be set to a higher value.
        # For the Brownian motion, the process_std directly gives the initial uncertainty on velocity
        initial_state = torch_kf.GaussianState(
            torch.zeros(len(unmatched_measures), self.kalman_filter.state_dim, 1, dtype=self.dtype),
            torch.stack([self.build_initial_covariance(detections.dim)] * (1 + len(unmatched_measures)))[1:],
        )
        initial_state.mean[:, : detections.dim, 0] = unmatched_measures
        initial_state.covariance[:, : detections.dim, : detections.dim] = self.kalman_filter.measurement_noise[
            : detections.dim, : detections.dim
        ]

        # Merge still active states with the initial ones of the created tracks
        self.active_states = torch_kf.GaussianState(
            torch.cat((self.active_states.mean[active_mask], initial_state.mean)),
            torch.cat((self.active_states.covariance[active_mask], initial_state.covariance)),
        )
        if self.specs.online_process_std:
            self.process_noises = torch.cat(
                (
                    self.process_noises[active_mask],
                    self.kalman_filter.process_noise[None].expand(
                        len(unmatched_measures), self.kalman_filter.state_dim, self.kalman_filter.state_dim
                    ),
                )
            )

        if self.specs.track_building == TrackBuilding.DETECTION:
            self.all_positions.append(
                torch.cat(
                    [
                        (
                            detections.position[track.detection_ids[-1]][None]
                            if track.detection_ids[-1] != -1
                            else torch.full((1, detections.position.shape[1]), torch.nan)
                        )
                        for track in self.active_tracks
                    ]
                )
            )
        else:
            self.all_positions.append(self.active_states.mean[:, : detections.dim, 0].to(torch.float32))

        if self.save_all or self.specs.track_building == TrackBuilding.SMOOTHED:
            self.all_states.append(self.active_states)



from abc import ABC, abstractmethod
from typing import Union

import numpy as np
import scipy.ndimage  # type: ignore


class OpticalFlow(ABC):
    """Optical flow algorithm (Abstract class, see implementations)

    It implements or wraps an existing optical flow method.

    In ByoTrack, frames are multidimensional arrays of shape ([D, ]H, W, C).
    Two frames are given as input to the algorithm. It outputs an optical flow map which
    is a multidimensional floating array (dim, [D, ]H, W) where the first dimension is the
    flow (displacement) for each spatial dimension.

    For instance, with two dimensional frames (H, W, C), the computed flow map has a shape
    (2, H, W), where the first (resp. second) element of the first dimension is the displacement
    along height (resp. width). The coordinate system of ByoTrack is using the pixel index (i, j) and not
    the pixel position (x, y) (which is often use in OpenCV). Note that in 3D, the pixel order is (k, i, j)
    where k stand for depth.

    Additionally, the wrapper allows to downscale (with a blurring) the image to speed up computations.

    Usage:

    .. code-block:: python

        video: Video
        optflow: OpticalFlow
        detector: BatchDetector

        # Preprocess the frames (reference and moving)
        ref = optflow.preprocess(video[0])
        mov = optflow.preprocess(video[1])

        # Compute the flow map from ref to mov
        flow_map = optflow.compute(ref, mov)

        # flow_map[:, i, j] gives the displacement (di, dj) of the pixel (i, j) of the reference image (scaled)

        # You can warp mov into ref with warp
        mov_warp = optflow.warp(flow_map, video[1])

        # Or use the flow to move points, for instance positions of detected particles
        detections = detector.detect(video[0][None])[0]  # Detect with a BatchDetector on frame 0
        futur_positions = optflow.transform(detections.position.numpy())  # Expected positions on frame 1


    Attributes:
        downscale (Union[float, np.ndarray]): Downscale factor. Can be provided for each spatial axis.
            Default: 2.0
        blur (Union[None, float, np.ndarray]): Std of the Gaussian blur applied before downscaling the images
            Can be provided for each spatial axis. If not provided, a default one is choosen given the downscale.

    """

    def __init__(self, downscale: Union[float, np.ndarray] = 2.0, blur: Union[None, float, np.ndarray] = None):
        self.downscale = downscale
        self.blur = blur

    def preprocess(self, frame: np.ndarray) -> np.ndarray:
        """Downscale a frame before running the optical flow algorithm

        It blurs then downscales the frame. It also converts to float32.

        Args:
            frame (np.ndarray): The frame to pre-preprocess
                It does not normalize the frame. Integer values are simply converted to floating ones.
                Shape: ([D, ]H, W, C)

        Returns:
            np.ndarray: The blurred and downscaled frame (converted to float32)
                Shape: ([D', ]H', W', C), dtype: float32

        """
        downscale = np.ones(frame.ndim, np.float32)
        downscale[:-1] = self.downscale  # Do not downscale channels

        blur = np.zeros(frame.ndim, np.float32)  # Do not blur along channel dimension
        if self.blur is None:
            blur[:-1] = 2 * downscale[:-1] / 6
        else:
            blur[:-1] = self.blur

        # Blur
        frame = scipy.ndimage.gaussian_filter(frame.astype(np.float32), blur, mode="reflect")

        # Downscale (Linear interpolation)
        return scipy.ndimage.zoom(frame, 1 / downscale, order=1, mode="mirror", grid_mode=True)

    @abstractmethod
    def compute(self, reference: np.ndarray, moving: np.ndarray) -> np.ndarray:
        """Compute the optical flow map from reference to the moving frame

        It computes the displacement of each pixel from the reference frame to be in the moving one.

        Args:
            reference (np.ndarray): Reference frame
                Shape: ([D', ]H', W', C), dtype: float
            moving (np.ndarray): Moving frame
                Shape: ([D', ]H', W', C), dtype: float

        Returns:
            np.ndarray: Optical flow map from reference to moving
                The flow field is stored in pixel coordinates ([dk', ]di', dj') (!= xyz)
                Shape: (dim, [D', ]H', W'), dtype: float32
        """

    def flow_at(self, flow_map: np.ndarray, points: np.ndarray) -> np.ndarray:
        """Extract the flow/displacement at the given location

        The flow_map is expected to be downscaled, but the input and outputs points are not.

        Args:
            flow_map (np.ndarray): Optical flow map. Displacement ([dk', ]di', dj') for each pixel ([k', ]i', j') in
                the downscaled coordinates.
                Shape: (dim, [D', ]H', W'), dtype: float32
            points (np.ndarray): Points ([k, ]i ,j) in the original (non-downscale) coordinates
                Shape: (N, dim), dtype: float

        Returns:
            np.ndarray: The displacements ([dk, ], di, dj) in the original coordinates (non-downscale)
                at the given points.
                Shape: (N, dim), dtype: float
        """
        if isinstance(self.downscale, (int, float)):
            downscale = np.array([self.downscale] * flow_map.shape[0])
        else:
            downscale = self.downscale

        flow = np.empty_like(points)

        for axis in range(flow_map.shape[0]):
            flow[:, axis] = scipy.ndimage.map_coordinates(flow_map[axis], points.T / downscale[axis], mode="nearest")

        flow *= downscale

        return flow

    def transform(self, flow_map: np.ndarray, points: np.ndarray) -> np.ndarray:
        """Apply the flow to the given points

        The flow_map is expected to be downscaled, but the input and outputs points are not.

        Args:
            flow_map (np.ndarray): Optical flow map. Displacement ([dk', ]di', dj') for each pixel ([k', ]i', j') in
                the downscaled coordinates.
                Shape: (dim, [D', ]H', W'), dtype: float32
            points (np.ndarray): Points ([k, ]i ,j) in the original (non-downscale) coordinates
                Shape: (N, dim), dtype: float

        Returns:
            np.ndarray: New positions of the poitns ([k, ]i, j) in the original (non-downscale) coordinates
                once the flow is applied.
                Shape: (N, dim), dtype: float
        """
        return points + self.flow_at(flow_map, points)

    def warp(self, flow_map: np.ndarray, moving: np.ndarray) -> np.ndarray:
        """Warps the moving image onto the reference using the flow map

        It warps the non-preprocessed moving image (in the original non-downscaled coordinates).

        Note:
            We only implemented backward warping (It only allows to warp the moving image onto the reference one)
            Forward warping could be implemented in a future version if needed.

        Args:
            flow_map (np.ndarray): Optical flow map. Displacement ([dk', ]di', dj') for each pixel ([k', ]i', j') in
                the downscaled coordinates.
                Shape: (dim, [D', ]H', W'), dtype: float32
            moving (np.ndarray): The moving image to warp (non-preprocessed)
                Shape: ([D, ]H, W, C), dtype: float

        Returns:
            np.ndarray: Warped image
                Shape: ([D, ]H, W, C), dtype: float

        """
        *dims, channels = moving.shape
        points: np.ndarray = np.indices(dims, np.float32).reshape(len(dims), -1).transpose()
        points = self.transform(flow_map, points).transpose().reshape(len(dims), *dims)

        return np.concatenate(
            [
                scipy.ndimage.map_coordinates(moving[..., channel], points, mode="nearest", order=1)[..., None]
                for channel in range(channels)
            ],
            axis=-1,
        )


class DummyOpticalFlow(OpticalFlow):
    """Dummy optical flow which predict no displacement for each pixel"""

    def compute(self, reference: np.ndarray, moving: np.ndarray) -> np.ndarray:
        return np.zeros((reference.ndim - 1, *reference.shape[:-1]))



from typing import Collection, List, Sequence, Tuple, Union
import warnings

import numpy as np
import torch
import torch_kf
import torch_kf.ckf
import tqdm.auto as tqdm

import byotrack


# TODO: Add a KOFTSmoother ? that uses optical flow to smooth tracks as in KOFT
# TODO: Support for float64 and cuda ?


class RTSSmoother(byotrack.Refiner):  # pylint: disable=too-few-public-methods
    """Smooth tracks positions by running the optimal Rauch-Tung-Striebel smoothing.

    It will also fill out NaN values in tracks. (Except if a track starts with NaN values,
    the track will now starts at the first non-NaN value.)

    It models the tracks motion and position measurement using a Kalman filter framework (with Gaussian iid noise).
    In order to have consistent results, the tracks positions should be extracted as the positions of an independant
    detection process that is not already filtered/smoothed.

    See `torch_kf` and `filterpy` for more details.

    Note:
        This implementation requires torch-kf. (pip install torch-kf)

    Attributes:
        detection_std (Union[float, torch.Tensor]): Expected measurement noise on the detection process.
            The detection process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension). See `torch_kf.ckf.constant_kalman_filter`.
            Default: 3.0 pixels
        process_std (Union[float, torch.Tensor]): Expected process noise. See `torch_kf.ckf.constant_kalman_filter`, the
            process is modeled as constant order-th derivative motion. This quantify how much the supposely "constant"
            order-th derivative can change between two consecutive frames. A common rule of thumb is to use
            3 * process_std ~= max_t(| x^(order)(t) - x^(order)(t+1)|). It can be provided for each dimension).
            Default: 1.5 pixels / frame^order
        kalman_order (int): Order of the Kalman filter to use.
            0 for brownian motions, 1 for directed brownian motion, 2 for accelerated brownian motions, etc...
            Default: 1
        anisotropy (Tuple[float, float, float]): Deprecated and ignored. If the data is anisotrope, it should directly
            be incorporated inside detection_std and process_std. Will be removed in a future version.
        initial_std_factor (float): The initial state is created with an initial uncertainty
            set to the measurement_std of the Kalman Filter on measured states and this factor times
            the process_std for unmeasured states.
            Default: 10.0

    """

    description_forward = "RTS smoothing (forward)"
    description_backward = "RTS smoothing (backward)"

    def __init__(
        self,
        detection_std: Union[float, torch.Tensor] = 3.0,
        process_std: Union[float, torch.Tensor] = 1.5,
        kalman_order: int = 1,
        *,
        anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0),
        initial_std_factor=10.0,
    ) -> None:
        super().__init__()
        self.detection_std = detection_std
        self.process_std = process_std
        self.kalman_order = kalman_order
        self.initial_std_factor = initial_std_factor

        if anisotropy != (1.0, 1.0, 1.0):
            warnings.warn(
                "'anisotropy' is deprecated and will be removed in a future version; it is ignored.",
                DeprecationWarning,
                stacklevel=2,
            )

    def run(  # pylint: disable=too-many-locals
        self, _: Union[Sequence[np.ndarray], np.ndarray], tracks: Collection[byotrack.Track]
    ) -> List[byotrack.Track]:
        if not tracks:
            return []

        dim = next(iter(tracks)).points.shape[-1]

        kalman_filter = torch_kf.ckf.constant_kalman_filter(
            self.detection_std,
            self.process_std,
            dim=dim,
            order=self.kalman_order,
        )

        # Starting time reference (positions[0] correspond to positions at start)
        start = min(track.start for track in tracks)

        positions = byotrack.Track.tensorize(tracks)
        is_defined = torch.full((positions.shape[0], positions.shape[1]), False)
        offsets: List[int] = []

        for i, track in enumerate(tracks):
            offset = int(torch.nonzero(~torch.isnan(track.points).any(dim=-1))[0, 0].item())
            if offset != 0:  # Handle tracks starting by NaN values
                warnings.warn("A track is starting with NaN values. It will be clipped.")

            is_defined[track.start + offset - start : track.start + len(track) - start, i] = True
            offsets.append(offset)

        # Initial state (see KalmanLinker)
        process_std = torch.broadcast_to(torch.as_tensor(self.process_std), (dim,)) * self.initial_std_factor
        process_std = torch.cat([process_std**2] * (kalman_filter.state_dim // dim))  # Expand to state_dim

        initial_state = torch_kf.GaussianState(
            torch.zeros(1, kalman_filter.state_dim, 1),
            torch.diag(process_std)[None],
        )
        initial_state.covariance[:, :dim, :dim] = kalman_filter.measurement_noise

        # Filtering
        states = torch_kf.GaussianState(
            torch.zeros(positions.shape[0], positions.shape[1], kalman_filter.state_dim, 1),
            torch.zeros(positions.shape[0], positions.shape[1], kalman_filter.state_dim, kalman_filter.state_dim),
        )

        was_defined = torch.full((positions.shape[1],), False)
        for frame_id in tqdm.trange(positions.shape[0], desc=self.description_forward):
            starting = ~was_defined & is_defined[frame_id]
            kept = was_defined & is_defined[frame_id]
            measured = ~torch.isnan(positions[frame_id]).any(dim=-1)

            # Prediction
            if frame_id > 0:
                states[frame_id, kept] = kalman_filter.predict(states[frame_id - 1, kept])

            # Initialization
            assert measured[starting].all()  # Always measured on start point
            states[frame_id, starting] = initial_state
            states.mean[frame_id, starting, :dim, 0] = positions[frame_id, starting]

            # Update
            states[frame_id, kept & measured] = kalman_filter.update(
                states[frame_id, kept & measured], positions[frame_id, kept & measured, :, None]
            )

            was_defined = is_defined[frame_id]

        # Smoothing: iterate backward (Update done for active t where t+1 is defined)
        for frame_id in tqdm.trange(positions.shape[0] - 2, -1, -1, desc=self.description_backward):
            kept = is_defined[frame_id + 1] & is_defined[frame_id]
            cov_at_process = states.covariance[frame_id, kept] @ kalman_filter.process_matrix.mT
            predicted_covariance = kalman_filter.process_matrix @ cov_at_process + kalman_filter.process_noise

            kalman_gain = cov_at_process @ predicted_covariance.inverse().mT
            states.mean[frame_id, kept] += kalman_gain @ (
                states.mean[frame_id + 1, kept] - kalman_filter.process_matrix @ states.mean[frame_id, kept]
            )
            states.covariance[frame_id, kept] += (
                kalman_gain @ (states.covariance[frame_id + 1, kept] - predicted_covariance) @ kalman_gain.mT
            )

        # Build new tracks from computed states
        new_tracks = []
        for i, track in enumerate(tracks):
            new_tracks.append(
                byotrack.Track(
                    track.start,
                    states.mean[track.start + offsets[i] - start : track.start + len(track) - start, i, :dim, 0],
                    track.identifier,
                    track.detection_ids[offsets[i] :],
                    merge_id=track.merge_id,
                    parent_id=track.parent_id,
                )
            )

        # Check produced tracks
        byotrack.Track.check_tracks(new_tracks, warn=True)

        return new_tracks



import dataclasses
import enum
from typing import List, Optional, Tuple, Union
import warnings

import numpy as np
import torch
import torch_kf
import torch_kf.ckf

import byotrack

from .nearest_neighbor import AssociationMethod, FrameByFrameLinker, FrameByFrameLinkerParameters


class Cost(enum.Enum):
    """The cost to use for association

    * LIKELIHOOD
        Build the cost matrix to maximize the likelihood of the association.
        The cost is defined as: the negative log likelihood that the detections j is from track i.
        C[i, j] = - log P(det_j | track_i)
        The cost limit is expected to be the smallest probability to accept.
    * MAHALANOBIS
        Mahalanobis distance, using the track uncertainty to correct the euclidean distance.
        The cost limit is expected to be the largest Mahalanobis distance to accept.
    * MAHALANOBIS_SQ
        Squared Mahalanobis distance.
        The cost limit is expected to be the largest Mahalanobis distance to accept. (not squared)
    * EUCLIDEAN
        Euclidean distance.
        The cost limit is expected to be the largest Euclidean distance to accept.
    * EUCLIDEAN_SQ
        Squared euclidean distance.
        The cost limit is expected to be the largest Euclidean distance to accept. (not squared)

    """

    LIKELIHOOD = "likelihood"
    MAHALANOBIS = "mahalanobis"
    MAHALANOBIS_SQ = "mahalanobis_sq"
    EUCLIDEAN = "euclidean"
    EUCLIDEAN_SQ = "euclidean_sq"


class TrackBuilding(enum.Enum):
    """How to build the final tracks

    * DETECTION
        Build tracks from detections without filtering nor filling gaps
    * FILTERED
        Build tracks from the Kalman filter outputs
    * SMOOTHED
        Add a backward run of the Kalman filter (RTS) to smooth tracks position

    """

    DETECTION = "detection"
    FILTERED = "filtered"
    SMOOTHED = "smoothed"


@dataclasses.dataclass
class KalmanLinkerParameters(FrameByFrameLinkerParameters):
    """Parameters of KalmanLinker

    Note:
        The merging and splitting features is still experimental.

    Attributes:
        association_threshold (float): This is the main hyperparameter, it defines the threshold on the distance used
            not to link tracks with detections. It prevents to link with false positive detections.
            Default: 5 pixels
        detection_std (Union[float, torch.Tensor]): Expected measurement noise on the detection process.
            The detection process is modeled with a Gaussian noise with this given std. (You can provide a different
            noise for each dimension). See `torch_kf.ckf.constant_kalman_filter`.
            Default: 3.0 pixels
        process_std (Union[float, torch.Tensor]): Expected process noise. See `torch_kf.ckf.constant_kalman_filter`, the
            process is modeled as constant order-th derivative motion. This quantify how much the supposely "constant"
            order-th derivative can change between two consecutive frames. A common rule of thumb is to use
            3 * process_std ~= max_t(| dx^(order)(t+1) - dx^(order)(t)|). It can be provided for each dimension).
            Default: 1.5 pixels
        kalman_order (int): Order of the Kalman filter to use.
            0 for brownian motions, 1 for directed brownian motion, 2 for accelerated brownian motions, etc...
            Default: 1
        n_valid (int): Number associated detections required to validate the track after its creation.
            Default: 3
        n_gap (int): Number of consecutive frames without association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, [SPARSE_]OPT_HARD, [SPARSE_]OPT_SMOOTH)
            Default: OPT_SMOOTH
        anisotropy (Tuple[float, float, float]): Anisotropy of images (Ratio of the pixel sizes
            for each axis, depth first). This will be used to scale distances. It will only impact
            EUCLIDEAN[_SQ] costs. For probabilistic cost, anisotropy should be already integrated
            in the stds of the kalman filter (providing one std for each dimension).
            Default: (1., 1., 1.)
        cost_method (CostMethod): The cost method to use. It can be provided as a string.
            See `CostMethod`. It also indicates what is the correct unit of `association_threshold`.
            Default: EUCLIDEAN
        track_building (TrackBuilding): Tells the linker how to build the final tracks.
            Either from detections, or from filtered/smoothed positions computed by the
            Kalman filter. See `TrackBuilding`. It can be provided as a string.
            Default: FILTERED
        split_factor (float): Allow splitting of tracks, using a second association step.
            The association threshold in this case is `split_factor * association_threshold`.
            Default: 0.0 (No splits)
        merge_factor (float): Allow merging of tracks, using a second association step.
            The association threshold in this case is `merge_factor * association_threshold`.
            Default: 0.0 (No merges)
        online_process_std (float): Recomputes the process std online following "A. Genovesio, et al, 2004, October.
            Adaptive gating in Gaussian Bayesian multi-target tracking. ICIP'04. (Vol. 1, pp. 147-150). IEEE."
            Each track has its own process std depending on the errors made in the past. It automatically adjusts to
            process errors, allowing to increase the validation gate. Should be used in conjonction with MAHALANOBIS
            or LIKELIHOOD `cost_method`. As this may be detrimental, it is disabed by default.
            Default: 0.0 (Process_std is constant)
        initial_std_factor (float): The uncertainties on initial velocities/accelerations are set
            to initial_std_factor * process_std. See `KalmanLinker.build_initial_covariance`.
            Having a small factor will prevent handling correctly starting tracks with large initial velocity
            on their first frames. But large values will lead to large uncertainty on the first prediction, making
            it hard to associate to a detection with MAHALANOBIS or LIKELIHOOD methods.
            Typical values lies between 3.0 to 10.0.
            Default: 10.0

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        association_threshold: float = 5.0,
        *,
        detection_std: Union[float, torch.Tensor] = 3.0,
        process_std: Union[float, torch.Tensor] = 1.5,
        kalman_order: int = 1,
        n_valid=3,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0),
        cost: Union[str, Cost] = Cost.EUCLIDEAN,
        track_building: Union[str, TrackBuilding] = TrackBuilding.FILTERED,
        split_factor: float = 0.0,
        merge_factor: float = 0.0,
        online_process_std: float = 0.0,
        initial_std_factor: float = 10.0,
    ):
        super().__init__(
            association_threshold=association_threshold,
            n_valid=n_valid,
            n_gap=n_gap,
            anisotropy=anisotropy,
            association_method=association_method,
            split_factor=split_factor,
            merge_factor=merge_factor,
        )

        if isinstance(detection_std, float) and min(anisotropy) != max(anisotropy):
            warnings.warn(
                "A single `detection_std` is provided, but images are anisotrope. Consider giving one std by dimension."
            )
        if isinstance(process_std, float) and min(anisotropy) != max(anisotropy):
            warnings.warn(
                "A single `process_std` is provided, but images are anisotrope. Consider giving one std by dimension."
            )

        self.detection_std = detection_std
        self.process_std = process_std
        self.kalman_order = kalman_order
        self.initial_std_factor = initial_std_factor

        self.cost = cost if isinstance(cost, Cost) else Cost[cost.upper()]
        self.track_building = (
            track_building if isinstance(track_building, TrackBuilding) else TrackBuilding[track_building.upper()]
        )
        self.online_process_std = online_process_std

    detection_std: Union[float, torch.Tensor] = 3.0
    process_std: Union[float, torch.Tensor] = 1.5
    kalman_order: int = 1
    cost: Cost = Cost.EUCLIDEAN
    track_building: TrackBuilding = TrackBuilding.FILTERED
    online_process_std: float = 0.0
    initial_std_factor: float = 10.0


class KalmanLinker(FrameByFrameLinker):
    """Frame by frame linker using Kalman filters

    Motion is modeled with a Kalman filter of a specified order (See `torch_kf.ckf`)
    Matching is done to optimize the given cost. If optical flow is provided, it is used
    online to warp the predicted state positions of the kalman filter. This will work, but it
    is sub-optimal: consider using `KOFTLinker` that exploits in a finer way optical flow
    inside Kalman filters.

    This is an implementation of Simple Kalman Tracking (SKT) from KOFT [9].

    Note:
        This implementation requires torch-kf. (pip install torch-kf)

    See `FrameByFrameLinker` for the other attributes.

    Attributes:
        specs (KalmanLinkerParameters): Parameters specifications of the algorithm.
            See `KalmanLinkerParameters`.
        kalman_filter (Optional[torch_kf.KalmanFilter]): The Kalman filter. (Build once the tracking starts
            allowing to adapt the dimension on the fly)
        active_states (Optional[torch_kf.GaussianState]): The Kalman filter estimation for each track.
            Shape: mean=(N, D * (order + 1), 1), covariance=(N, D * (order + 1), dim * (order + 1))
            dtype: float
        projections (Optional[torch_kf.GaussianState]): The Kalman filter projection for each track.
            Shape: mean=(N, D, 1), covariance=(N, D, D), precision=(N, D, D)
            dtype: float
        process_noises (Optional[torch.Tensor]): The Kalman filter process noise for each track.
            Only used when online_process_std > 0.0. It allows to compute an adaptative process_std
            and therefore gating for each track.
            Shape: (N, D, 1), dtype: float
        all_states (List[torch_kf.GaussianState]): The Kalman filter estimation for each track at each seen
            frame. States are only registered when save_all=True or if you build tracks from RTS smoothing.
            Shape: mean=(N, D * (order + 1), 1), covariance=(N, D * (order + 1), dim * (order + 1))
            dtype: float

    """

    progress_bar_description = "Kalman filter linking"

    def __init__(
        self,
        specs: KalmanLinkerParameters,
        optflow: Optional[byotrack.OpticalFlow] = None,
        features_extractor: Optional[byotrack.FeaturesExtractor] = None,
        save_all=False,
    ) -> None:
        super().__init__(specs, optflow, features_extractor, save_all)

        self.specs: KalmanLinkerParameters
        self.kalman_filter = torch_kf.ckf.constant_kalman_filter(
            self.specs.detection_std,
            self.specs.process_std,
            dim=2,
            order=self.specs.kalman_order,
        )
        self.dtype = torch.float32

        self.active_states = torch_kf.GaussianState(
            torch.empty((0, self.kalman_filter.state_dim, 1)),
            torch.empty((0, self.kalman_filter.state_dim, self.kalman_filter.state_dim)),
        )
        self.projections = self.kalman_filter.project(self.active_states)
        self.process_noises = torch.empty((0, self.kalman_filter.state_dim, self.kalman_filter.state_dim))

        self.all_states: List[torch_kf.GaussianState] = []

    def reset(self, dim=2) -> None:
        super().reset(dim)

        self.kalman_filter = torch_kf.ckf.constant_kalman_filter(
            self.specs.detection_std,
            self.specs.process_std,
            dim=dim,
            order=self.specs.kalman_order,
            approximate=True,
        )
        self.kalman_filter = self.kalman_filter.to(self.dtype)
        self.active_states = torch_kf.GaussianState(
            torch.empty((0, self.kalman_filter.state_dim, 1), dtype=self.dtype),
            torch.empty((0, self.kalman_filter.state_dim, self.kalman_filter.state_dim), dtype=self.dtype),
        )
        self.projections = self.kalman_filter.project(self.active_states)
        self.process_noises = torch.empty(
            (0, self.kalman_filter.state_dim, self.kalman_filter.state_dim), dtype=self.dtype
        )
        self.all_states = []

    def collect(self) -> List[byotrack.Track]:
        if self.specs.track_building == TrackBuilding.SMOOTHED:
            dim = self.kalman_filter.state_dim
            tracks_handlers = [
                handler
                for handler in self.inactive_tracks + self.active_tracks
                if handler.track_state in (handler.TrackState.VALID, handler.TrackState.FINISHED)
            ]

            states = torch_kf.GaussianState(
                torch.full((len(self.all_states), len(tracks_handlers), dim, 1), torch.nan, dtype=self.dtype),
                torch.zeros((len(self.all_states), len(tracks_handlers), dim, dim), dtype=self.dtype),
            )
            is_defined = torch.full((len(self.all_states), len(tracks_handlers)), False)

            for t, states_ in enumerate(self.all_states):
                for i, handler in enumerate(tracks_handlers):
                    time_offset = t - handler.start
                    if 0 <= time_offset < len(handler):
                        states[t, i] = states_[handler.track_ids[time_offset]]
                        is_defined[t, i] = True

            # Iterate backward to update all states (Update done for active t where t+1 is defined)
            for t in range(len(self.all_states) - 2, -1, -1):
                mask = is_defined[t + 1] & is_defined[t]
                cov_at_process = states.covariance[t, mask] @ self.kalman_filter.process_matrix.mT
                predicted_covariance = (
                    self.kalman_filter.process_matrix @ cov_at_process + self.kalman_filter.process_noise
                )

                kalman_gain = cov_at_process @ predicted_covariance.inverse().mT
                states.mean[t, mask] += kalman_gain @ (
                    states.mean[t + 1, mask] - self.kalman_filter.process_matrix @ states.mean[t, mask]
                )
                states.covariance[t, mask] += (
                    kalman_gain @ (states.covariance[t + 1, mask] - predicted_covariance) @ kalman_gain.mT
                )

            dim = self.all_positions[0].shape[-1]  # For KOFT, using kf.measure_dim would not work
            tracks = []
            for i, handler in enumerate(tracks_handlers):
                tracks.append(
                    byotrack.Track(
                        handler.start,
                        states.mean[handler.start : handler.start + len(handler), i, :dim, 0].to(torch.float32),
                        handler.identifier,
                        torch.tensor(handler.detection_ids[: len(handler)], dtype=torch.int32),
                        merge_id=handler.merge_id,
                        parent_id=handler.parent_id,
                    )
                )

            return tracks

        return super().collect()

    def motion_model(self) -> None:
        # Use the Kalman filter to predict the current states of each active tracks
        self.active_states = self.kalman_filter.predict(
            self.active_states, process_noise=self.process_noises if self.specs.online_process_std else None
        )

        # Add optical flow motion to the position
        if self.optflow and self.optflow.flow_map is not None:
            positions = self.active_states.mean[:, : self.kalman_filter.measure_dim, 0]
            positions[:] = torch.tensor(self.optflow.optflow.transform(self.optflow.flow_map, positions.numpy()))

        # Project states for association
        self.projections = self.kalman_filter.project(self.active_states)

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        anisotropy = torch.tensor(self.specs.anisotropy)[-detections.dim :]

        if self.specs.cost == Cost.EUCLIDEAN:
            return (
                torch.cdist(self.projections.mean[..., 0] * anisotropy, detections.position * anisotropy),
                self.specs.association_threshold,
            )

        if self.specs.cost == Cost.EUCLIDEAN_SQ:
            return (
                torch.cdist(self.projections.mean[..., 0] * anisotropy, detections.position * anisotropy).pow_(2),
                self.specs.association_threshold**2,
            )
        if self.specs.cost == Cost.MAHALANOBIS:
            return (
                self.projections[:, None].mahalanobis(detections.position[None, ..., None]),
                self.specs.association_threshold,
            )
        if self.specs.cost == Cost.MAHALANOBIS_SQ:
            return (
                self.projections[:, None].mahalanobis_squared(detections.position[None, ..., None]),
                self.specs.association_threshold**2,
            )

        # LIKELIHOOD: cost = -log likelihood
        cost = -self.projections[:, None].log_likelihood(detections.position[None, ..., None])
        return cost, -torch.log(torch.tensor(self.specs.association_threshold)).item()

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, active_mask: torch.Tensor):
        positions = detections.position.to(self.dtype)

        # Update the state of associated tracks (unassociated tracks keep the predicted state)
        if self.specs.online_process_std:
            updated = self.kalman_filter.update(
                self.active_states[self._links[:, 0]],
                positions[self._links[:, 1]][..., None],
                projection=self.projections[self._links[:, 0]],
            )

            # Update the process_noise based on the prediction errors
            # Err = x_t - Fx_{t-1} (low estimation of errors)
            # Then Q_t = a * (b * Q_t + 1 - b err @ err.T) + (1- a) Q_0
            # In this implem, we fix b = 0.75 (around 4 frames to estimate errors) and a is user-defined.
            errors = self.active_states.mean[self._links[:, 0]] - updated.mean
            self.process_noises[self._links[:, 0]] *= self.specs.online_process_std * 0.75  # Fixed EMA ~4 frames lag
            self.process_noises[self._links[:, 0]] += self.specs.online_process_std * 0.25 * errors @ errors.mT
            self.process_noises[self._links[:, 0]] += (
                1 - self.specs.online_process_std
            ) * self.kalman_filter.process_noise

            self.active_states[self._links[:, 0]] = updated
        else:
            self.active_states[self._links[:, 0]] = self.kalman_filter.update(
                self.active_states[self._links[:, 0]],
                positions[self._links[:, 1]][..., None],
                projection=self.projections[self._links[:, 0]],
            )

        # Create new states for unmatched measures
        unmatched_measures = positions[self._unmatched_detections]

        # Build the initial states for tracks:
        # Infinite uncertainty on the prior position => Given positional measurement, we initialize on its position
        #                                               with the positional measurement uncertainty.
        # For derivatives, we assume they are centered on 0 (no bias toward a direction) with an uncertainty of
        # initial_std_factor * process_std:
        # Small factors (~1) are not suited for tracking objects that appears with a large initial velocity.
        # Large factors (>> 1) will create association problem on the second frame, as the predicted
        # position inherit from this uncertainty.
        initial_state = torch_kf.GaussianState(
            torch.zeros(len(unmatched_measures), self.kalman_filter.state_dim, 1, dtype=self.dtype),
            torch.stack([self.build_initial_covariance(detections.dim)] * (1 + len(unmatched_measures)))[1:],
        )
        initial_state.mean[:, : detections.dim, 0] = unmatched_measures
        initial_state.covariance[:, : detections.dim, : detections.dim] = self.kalman_filter.measurement_noise[
            : detections.dim, : detections.dim
        ]

        # Merge still active states with the initial ones of the created tracks
        self.active_states = torch_kf.GaussianState(
            torch.cat((self.active_states.mean[active_mask], initial_state.mean)),
            torch.cat((self.active_states.covariance[active_mask], initial_state.covariance)),
        )
        if self.specs.online_process_std:
            self.process_noises = torch.cat(
                (
                    self.process_noises[active_mask],
                    self.kalman_filter.process_noise[None].expand(
                        len(unmatched_measures), self.kalman_filter.state_dim, self.kalman_filter.state_dim
                    ),
                )
            )

        if self.specs.track_building == TrackBuilding.DETECTION:
            self.all_positions.append(
                torch.cat(
                    [
                        (
                            detections.position[track.detection_ids[-1]][None]
                            if track.detection_ids[-1] != -1
                            else torch.full((1, detections.position.shape[1]), torch.nan)
                        )
                        for track in self.active_tracks
                    ]
                )
            )
        else:
            self.all_positions.append(self.active_states.mean[:, : detections.dim, 0].to(torch.float32))

        if self.save_all or self.specs.track_building == TrackBuilding.SMOOTHED:
            self.all_states.append(self.active_states)

    def build_initial_covariance(self, dim: int) -> torch.Tensor:
        """Build the diagonal initial covariance matrix

        The position is initially unknown, leading to a belief (given by the first detection) set
        to the position of the first detection, with detection_std uncertainty.

        The velocity (and higher order derivatives) are assumed to be 0.0 with a relatively high uncertainty:
        initial_std_factor * process_std.

        Note that having a large initial_std_factor (>10) may decrease performances, as the first prediction
        will be impacted and largely uncertain, leading to low probabilies for every associations. In KOFT,
        as the velocity is measured before the first prediction, the initial_std_factor can be increased to
        reduce this bias toward a nul initial velocity. We found that initial_std_factor=0.0 is a good
        trade off in practice.
        """
        process_std = torch.broadcast_to(torch.as_tensor(self.specs.process_std, dtype=self.dtype), (dim,)).clone()

        # In the case of Brownian motion, the intitial covariance if fully rewritten in SKT (no impact)
        # But in KOFT, Brownian motion models velocity with an initial velocity centered on 0 and with
        # an uncertainty given by the process_std, therefore we don't use initial_std_factor
        if self.specs.kalman_order > 0:
            process_std *= self.specs.initial_std_factor

        # Process std is squared and set for each order of the process (pos, vel, acc, ...)
        covariance = torch.diag(torch.cat([process_std**2] * (self.kalman_filter.state_dim // dim)))

        # Then, it is overwritten for the position, to be set to measurement_std
        measurement_std = torch.broadcast_to(torch.as_tensor(self.specs.detection_std, dtype=self.dtype), (dim,))
        torch.diagonal(covariance)[:dim] = measurement_std**2

        return covariance



import re
from typing import Iterable, List, TypeVar

T = TypeVar("T")


def sorted_alphanumeric(data: Iterable[T]) -> List[T]:
    """Sorts alphanumeriacally an iterable of strings-like objects

    "1" < "2" < "10" < "foo1" < "foo2" < "foo3"
    """

    def convert(text: str):
        return int(text) if text.isdigit() else text.lower()

    def alphanum_key(key):
        return tuple(convert(c) for c in re.split("([0-9]+)", str(key)))

    return sorted(data, key=alphanum_key)



import time
from typing import cast, List, Optional, Union, Type
import warnings

import numba  # type: ignore
import numpy as np
import scipy.ndimage as ndi  # type: ignore
import torch

import byotrack


class B3SplineUWT(torch.nn.Module):
    """Undecimated Wavelet Transform with B3Spline for nD images.

    Also called A trous Wavelet Transform.

    Let J be the level of the UWD.
    Let c_0 be the original image. It computes successively dilated convolution of it.

    c_j = h_j * c_{j-1}, 1 \\le j \\le J
    w_j = c_{j-1} - c_{j}, 1 \\le j \\le J

    Where h_j is the filter at scale j which is the original filter h_1
    dilated with 2^{j - 1} 0 between each coefficient and h_1 = [1/16, 1/4, 3/8, 1/4, 1/16].

    The parameters returned are [w_1, w_2, ..., w_J, c_J], or only [w_J] if `return_all` is False
    The original image is easily reconstructed from the parameters with c_0 = c_J + \\sum_j w_j

    Boundaries issues are handled using mirror padding.

    """

    weights = torch.tensor([1 / 16, 1 / 4, 3 / 8, 1 / 4, 1 / 16])

    def __init__(self, level=3, dim=2, return_all=True):
        """Constructor

        Args:
            level (int): Level of the Wavelet Transform. (J)
            dim (int): Dimension of the UWT.
                Default: 2
            return_all (bool): Return all parameters and the remaining image (w_1, ... w_J, c_J)
                If False it only returns the parameters at the final level (w_J)
                Default: True

        """
        super().__init__()
        self.return_all = return_all
        self.dim = dim

        conv_builder: Union[Type[torch.nn.Conv1d], Type[torch.nn.Conv2d], Type[torch.nn.Conv3d]]
        if dim == 1:
            conv_builder = torch.nn.Conv1d
        elif dim == 2:
            conv_builder = torch.nn.Conv2d
        elif dim == 3:
            conv_builder = torch.nn.Conv3d
        else:
            raise NotImplementedError("B3SplineUWT are implemented only for 1d, 2d and 3d")

        self.layers = torch.nn.ModuleList(
            [
                conv_builder(1, 1, 5, dilation=2**i, padding=2 ** (i + 1), bias=False, padding_mode="reflect")
                for i in range(level)
            ]
        )

        # Construct multi dim weights
        weights = self.weights
        for _ in range(dim - 1):
            weights = weights[..., None] * self.weights

        for layer in self.layers:
            layer = cast(Union[torch.nn.Conv1d, torch.nn.Conv2d, torch.nn.Conv3d], layer)
            layer.weight.requires_grad_(False)
            layer.weight[0, 0] = weights

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the parameters of the UWD

        Args:
            x (torch.Tensor): Input multidimensional images without channel dimension
                Shape: (B, [D, ]H, W)

        Returns:
            torch.Tensor: Parameters [w_1, ..., w_J, c_J] or only w_J
                Shape: (B, J + 1, [D, ]H, W) or (B, [D, ]H, W) if `return_all` is False

        """
        batch, *spatial = x.shape
        assert len(spatial) == self.dim, f"The module was constructed for {self.dim}D images. Input is {len(spatial)}D."

        x = x[:, None]  # Add a channel dim: (B, 1, [D, ]H, W)
        y = x

        if self.return_all:
            outputs = torch.zeros((batch, len(self.layers) + 1, *spatial), device=x.device)

        for j, layer in enumerate(self.layers):
            x = y
            y = layer(y)

            if self.return_all:
                outputs[:, j] = (x - y)[:, 0]

        if self.return_all:
            outputs[:, -1] = y[:, 0]

            return outputs

        return (x - y)[:, 0]


class B3SplineUWTApprox1(torch.nn.Module):
    """Approximation of Undecimated Wavelet Transform with B3Spline for nD images.

    Split the nD convolution in n 1D convolution along each axis (as done in the original paper).
    Though it reduces the FLOPs it can lead to slower runtime (because of some torch optims).

    Depending on the gpu, the cudnn/cuda kernels, the pytorch version (and so on), it can sometimes
    be much faster than the nD convolution counterpart (especially in 3D).

    In this implementation, we rely on torch.nn.Conv1D, and is able to handle any number of dimension n.
    """

    weights = torch.tensor([1 / 16, 1 / 4, 3 / 8, 1 / 4, 1 / 16])

    def __init__(self, level=3, dim=2, return_all=True):
        """Constructor

        Args:
            level (int): Level of the Wavelet Transform. (J)
            dim (int): Dimension of the UWT.
                Default: 2
            return_all (bool): Return all parameters and the remaining image (w_1, ... w_J, c_J)
                If False it only returns the parameters at the final level (w_J)
                Default: True

        """
        super().__init__()
        self.return_all = return_all
        self.dim = dim

        self.layers = torch.nn.ModuleList(
            [
                torch.nn.Conv1d(1, 1, 5, dilation=2**i, padding=2 ** (i + 1), bias=False, padding_mode="reflect")
                for i in range(level)
            ]
        )

        for layer in self.layers:
            layer = cast(torch.nn.Conv1d, layer)
            layer.weight.requires_grad_(False)
            layer.weight[0, 0] = self.weights

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the parameters of the UWD

        Args:
            x (torch.Tensor): Input images without channel dimension
                Shape: (B, [D, ]H, W)

        Returns:
            torch.Tensor: Parameters [w_1, ..., w_J, c_J] or only [w_J]
                Shape: (B, J + 1, [D, ]H, W) or (B, [D, ]H, W) if `return_all` is False

        """
        batch, *spatial = x.shape
        assert len(spatial) == self.dim, f"The module was constructed for {self.dim}D images. Input is {len(spatial)}D."

        x = x[:, None]  # Add a channel dim: (B, 1, [D, ]H, W)
        y = x

        if self.return_all:
            outputs = torch.zeros((batch, len(self.layers) + 1, *spatial), device=x.device)

        for j, layer in enumerate(self.layers):
            x = y
            for axis in range(self.dim):
                # Transpose in ((B * S\{axis}, 1, S[axis])
                y = y.permute((0, *range(2, 2 + axis), *range(3 + axis, 2 + self.dim), 1, 2 + axis))
                y = layer(y.reshape(-1, 1, spatial[axis])).reshape(y.shape)
                # Transpose back
                y = y.permute((0, self.dim, *range(1, 1 + axis), 1 + self.dim, *range(1 + axis, self.dim)))

            if self.return_all:
                outputs[:, j] = (x - y)[:, 0]

        if self.return_all:
            outputs[:, -1] = y[:, 0]

            return outputs

        return (x - y)[:, 0]


class B3SplineUWTApprox2(torch.nn.Module):
    """Approximation of Undecimated Wavelet Transform with B3Spline for nD images.

    Split the nD convolution in n 1D convolution along each axis (as done in the original paper).
    Though it reduces the FLOPs it can lead to slower runtime (because of some torch optims).

    Depending on the gpu, the cudnn/cuda kernels, the pytorch version (and so on), it can sometimes
    be much faster than the nD convolution counterpart (especially in 3D).

    In this implementation, we rely on torch.nn.ConvnD with a reduced kernel.
    """

    weights = torch.tensor([1 / 16, 1 / 4, 3 / 8, 1 / 4, 1 / 16])

    def __init__(self, level=3, dim=2, return_all=True):
        """Constructor

        Args:
            level (int): Level of the Wavelet Transform. (J)
            dim (int): Dimension of the UWT.
                Default: 2
            return_all (bool): Return all parameters and the remaining image (w_1, ... w_J, c_J)
                If False it only returns the parameters at the final level (w_J)
                Default: True

        """
        super().__init__()
        self.return_all = return_all
        self.dim = dim

        conv_builder: Union[Type[torch.nn.Conv1d], Type[torch.nn.Conv2d], Type[torch.nn.Conv3d]]
        if dim == 1:
            conv_builder = torch.nn.Conv1d
        elif dim == 2:
            conv_builder = torch.nn.Conv2d
        elif dim == 3:
            conv_builder = torch.nn.Conv3d
        else:
            raise NotImplementedError("B3SplineUWT are implemented only for 1d, 2d and 3d")

        layers = []
        for i in range(level):
            sub_layers = []
            for j in range(dim):
                conv = conv_builder(
                    1,
                    1,
                    tuple(5 if k == j else 1 for k in range(dim)),  # type: ignore
                    dilation=tuple(2**i if k == j else 1 for k in range(dim)),  # type: ignore
                    padding=tuple(2 ** (i + 1) if k == j else 0 for k in range(dim)),  # type: ignore
                    bias=False,
                    padding_mode="reflect",
                )
                conv.weight.requires_grad_(False)
                conv.weight.view(5)[:] = self.weights
                sub_layers.append(conv)

            layers.append(torch.nn.Sequential(*sub_layers))

        self.layers = torch.nn.ModuleList(layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Compute the parameters of the UWD

        Args:
            x (torch.Tensor): Input images without channel dimension
                Shape: (B, [D, ]H, W)

        Returns:
            torch.Tensor: Parameters [w_1, ..., w_J, c_J] or only [w_J]
                Shape: (B, J + 1, [D, ]H, W) or (B, [D, ]H, W) if `return_all` is False

        """
        batch, *spatial = x.shape
        assert len(spatial) == self.dim, f"The module was constructed for {self.dim}D images. Input is {len(spatial)}D."

        x = x[:, None]  # Add a channel dim: (B, 1, [D, ]H, W)
        y = x

        if self.return_all:
            outputs = torch.zeros((batch, len(self.layers) + 1, *spatial), device=x.device)

        for j, layer in enumerate(self.layers):
            x = y
            y = layer(y)

            if self.return_all:
                outputs[:, j] = (x - y)[:, 0]

        if self.return_all:
            outputs[:, -1] = y[:, 0]

            return outputs

        return (x - y)[:, 0]


@numba.njit(cache=byotrack.NUMBA_CACHE)
def filter_small_objects(segmentation: np.ndarray, min_area: float) -> None:
    """Filter small instances from the segmentation in place

    Args:
        segmentation (np.ndarray): Segmentation mask to filtered inplace
            Shape ([D, ]H, W), dtype: integer
        min_area (float): Minimum number of pixels to be kept in the segmentation.

    """
    segmentation = segmentation.reshape(-1)
    area = np.zeros(segmentation.max(), np.uint)

    for i in range(segmentation.size):
        instance = segmentation[i] - 1
        if instance != -1:
            area[instance] += 1

    to_delete = area < min_area

    for i in range(segmentation.size):
        instance = segmentation[i] - 1
        if instance != -1:
            if to_delete[instance]:
                segmentation[i] = 0


class WaveletDetector(byotrack.BatchDetector):
    """Detection of bright spots using B3SplineUWT

    Following paper from Olivo-Marin, J.C. Extraction of spots in biological images using
    multiscale products. Pattern Recognit. 35, 1989-1996

    It supports 2D and 3D videos.

    The algorithm is in 4 steps:

    1. UWT decomposition
    2. Scale selection
    3. Noise filtering
    4. Connected components extraction

    The multi scales behavior (choosing multiple scales) was implemented but we decided to drop it.
    It adds complexity without real advantages from our experience.

    The same algorithm is implemented in Icy Software (SpotDetector). The main differences are:

    * nd wavelets (rather than n times one dimensional wavelets). It was designed to improve computations,
      but with torch no gain in time is observed in 2D. This can be switched either by calling `optimize`
      that will try to find the fastest option for your case, or manually by modifying the `b3swt` parameter.
    * Thresholding -> We follow the original paper using k times the std

    Attributes:
        scale (int): Scale of the wavelet coefficients used. With small scales, the detector focus on
            smaller objects.
        k (float): Noise threshold. Following the paper, the wavelet coefficients
            are filtered if coef \\le k \\sigma. (The higher the less spots you retrieve)
        min_area (float): Filter resulting spots that are too small (less than min_area pixels)
        device (torch.device): Device on which run the B3SplineUWT
            Default to cpu
        b3swt (B3SplineUWT): Undecimated wavelet transform
        **kwargs: Additional arguments for `BatchDetector` (batch_size, add_true_frames)

    """

    progress_bar_description = "Detections (Wavelet)"

    def __init__(self, scale=2, k=3.0, min_area=10.0, device: Optional[torch.device] = None, **kwargs):
        super().__init__(**kwargs)
        self.scale = scale
        self.k = k
        self.min_area = min_area
        self.device = device if device else torch.device("cpu")
        self.b3swt: Union[B3SplineUWT, B3SplineUWTApprox1, B3SplineUWTApprox2]
        self.b3swt = B3SplineUWT(scale + 1, return_all=False).to(self.device)

    def detect(self, batch: np.ndarray) -> List[byotrack.Detections]:
        assert (
            batch.shape[-1] == 1
        ), "This detector does not support multi channel images. Please aggregate the channels first"

        if batch.ndim - 2 != self.b3swt.dim:  # Wrong dim, let's reinit the module
            self.b3swt = type(self.b3swt)(self.scale + 1, dim=batch.ndim - 2, return_all=False).to(self.device)

        # Extract wavelet coefficient at the given scale
        inputs = torch.tensor(batch, dtype=torch.float32)[..., 0].to(self.device)
        coefficients: torch.Tensor = self.b3swt(inputs)

        # Thresholding. Shape: (B, [D, ]H, W)
        masks = (coefficients >= self.compute_threshold(coefficients)).cpu().numpy()

        # Instance segmentation by connected components
        detections_list = []
        for i, mask in enumerate(masks):
            segmentation = np.zeros(mask.shape, dtype=np.int32)
            ndi.label(mask, output=segmentation)  # 1-hop (4-ways in 2D) connected components by default

            if self.min_area:
                filter_small_objects(segmentation, self.min_area)

            detections_list.append(
                byotrack.Detections(
                    {
                        "segmentation": torch.tensor(segmentation),
                        # "confidence": put areas instead of deleting them ?
                    },
                    frame_id=i,
                )
            )

        return detections_list

    def compute_threshold(self, coefficients: torch.Tensor) -> torch.Tensor:
        """Compute threshold for the UWT coefficients

        Note: One could use MAD approx of \\sigma rather than std (or Icy implem) but it's almost equivalent
        (And k is a parameter to tune so it truly is)

        Args:
            coefficients (torch.Tensor): Coefficients of the UWT
                Shape: (B, [D, ]H, W)

        Returns:
            torch.Tensor: Threshold for each scale
                Shape: (B, 1, ..., 1)

        """
        return coefficients.std(dim=tuple(range(1, coefficients.ndim)), keepdim=True) * self.k

    def optimize(  # pylint: disable=too-many-statements,too-many-branches,too-many-locals
        self, frames: np.ndarray, repeat=5, warm_up=True, splines=(B3SplineUWT, B3SplineUWTApprox1, B3SplineUWTApprox2)
    ) -> "WaveletDetector":
        """Find the fastest configuration for the model on the given frames

        This is mainly designed for 3D videos, where 3D convolutions are heavy and poorly optimized with
        a dilation > 1 (for scales > 0). In particular, we observed cudnn kernels running 10 times faster
        on some gpus when determistic was set to False for 3D conv.

        epending on your hardware, kernels and pytorch version, one solution may be better than the other.
        This allows you to test most of the configuration and use the fastest one.

        Warning:
            With large 3D images, B3SplineUWTApprox1 may not work (it converts the spatial axis into batch axis,
            but conv1D do not support very large batch size. If you are in this case, you may disable it by
            changing the `splines` argument to `(B3SplineUWT, B3SplineUWTApprox2)`.

        Args:
            frames (np.ndarray): Frames of the video on which to test.
                Shape: (B, [D, ]H, W, C), dtype: float
            repeat (int): Number of time to repeat the computation to measure the timings.
                Default: 5
            warm_up (bool): Warm up each model before measuring time.
                Default: True
            splines (tuple): Implementation of B3SplineUWT to test. Reduce this list if one of the implementation
                is too long to run for instance.
                Default: (B3SplineUWT, B3SplineUWTApprox1, B3SplineUWTApprox2)

        Returns:
            WaveletDetector: self, with the best found b3swt. It may also modify pytorch backend for convolution.

        """
        inputs = torch.tensor(frames, dtype=torch.float32)[..., 0].to(self.device)
        models = [module(self.scale + 1, dim=inputs.ndim - 1, return_all=False).to(self.device) for module in splines]

        best_time = float("inf")
        best_model = models[0]

        if self.device.type == "cpu":
            for model in models:
                if warm_up:
                    model(inputs)

                total_time = 0.0
                for _ in range(repeat):
                    t = time.time()
                    model(inputs)
                    t = time.time() - t

                    if t > 5 * best_time:
                        # Stop early if it takes too long before the best one
                        break

                    total_time += t

                if t > 5 * best_time:
                    print(f"CPU: {model.__class__.__name__} computed in {t} => skip")
                    continue

                total_time /= repeat

                print(f"CPU: {model.__class__.__name__} compute in avg of {total_time}")

                if total_time < best_time:
                    best_time = total_time
                    best_model = model

        else:
            old_cudnn, old_deterministic = torch.backends.cudnn.enabled, torch.backends.cudnn.deterministic
            best_cudnn = old_cudnn
            best_deterministic = old_deterministic
            for cudnn, deterministic in [(True, True), (True, False), (False, False)]:
                torch.backends.cudnn.enabled = cudnn
                torch.backends.cudnn.deterministic = deterministic

                for model in models:
                    if warm_up:
                        model(inputs)

                    total_time = 0.0
                    for _ in range(repeat):
                        t = time.time()
                        model(inputs).cpu()
                        t = time.time() - t

                        if t > 5 * best_time:
                            # Stop early if it takes too long before the best one
                            break

                        total_time += t

                    if t > 5 * best_time:
                        print(
                            f"Cuda (cudnn: {cudnn}, deterministic: {deterministic})"
                            + f": {model.__class__.__name__} computed in {t} => skip"
                        )
                        continue

                    total_time /= repeat

                    print(
                        f"Cuda (cudnn: {cudnn}, deterministic: {deterministic})"
                        + f": {model.__class__.__name__} compute in avg of {total_time}"
                    )

                    if total_time < best_time:
                        best_time = total_time
                        best_model = model
                        best_cudnn = cudnn
                        best_deterministic = deterministic

            if best_cudnn and not old_cudnn:
                warnings.warn(f"Enabling cudnn backend (determinisitc={best_deterministic})")
            if not best_cudnn and old_cudnn:
                warnings.warn("Disabling cudnn backend")
            if old_deterministic != best_deterministic:
                warnings.warn(f"Switching cudnn deterministic to {deterministic}")

            torch.backends.cudnn.enabled = best_cudnn
            torch.backends.cudnn.deterministic = best_deterministic

        print(f"Selecting {best_model.__class__.__name__}")
        self.b3swt = best_model  # type: ignore

        return self



import dataclasses
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union
import warnings

import numpy as np
from packaging import version
import torch
import tqdm.auto as tqdm

# pylint: disable=import-error
import trackastra  # type: ignore
from trackastra.data import get_features, build_windows  # type: ignore
from trackastra.model import Trackastra, TrackingTransformer  # type: ignore
from trackastra.model.predict import predict_windows  # type: ignore
from trackastra.utils import normalize  # type: ignore

# pylint: enable=import-error


import byotrack
from .base import AssociationMethod, FrameByFrameLinker, FrameByFrameLinkerParameters


def build_cost_dict(
    nodes: List[Dict[str, Any]], weights: Iterable[Tuple[Tuple[int, int], float]]
) -> Dict[Tuple[int, int, int], Dict[int, float]]:
    """Build the cost dictionnary from Trackastra data format.

    It converts the feasible tracking graph predicted by trackastra into a mapping from edge to cost.
    Where an edge is a link between two detections (node). It converts the probability weight into a
    cost = -log weight.

    Args:
        nodes (List[Dict[str, Any]]): Nodes of the graph (detections), with their features
            Format: { "id": int, "coords": (float, float), "time": int, "label": int}
        weights (Iterable[Tuple[Tuple[int, int], float]]): Considered edges, with their weight
            Format: [((node_id, node_id_2), weight), ...]

    Returns:
        Dict[Tuple[int, int, int],Dict[int,float]]: Cost dictionnary of the feasible tracking graph.
        For every (frame1,node1,frame2) there is a dictionnary with key node2 and the value the cost of the edge.
    """
    cost_dict: Dict[Tuple[int, int, int], Dict[int, float]] = {}
    node_by_id = {node["id"]: node for node in nodes}
    for (id_1, id_2), weight in weights:
        node_1 = node_by_id[id_1]
        node_2 = node_by_id[id_2]
        key = (node_1["time"], node_1["label"] - 1, node_2["time"])
        if key not in cost_dict:
            cost_dict[key] = {}
        cost_dict[key][node_2["label"] - 1] = float(-np.log(weight))
    return cost_dict


class TrackastraFlex(Trackastra):  # pylint: disable=too-few-public-methods
    """Trackastra with ability to modify delta_t"""

    def __init__(
        self,
        transformer: TrackingTransformer,
        train_args: Dict[str, Any],
        delta_t: int = 4,
        intra_weight: float = 0,
        device=None,
    ):
        super().__init__(transformer, train_args, device)
        self.delta_t = delta_t
        self.intra_weight = intra_weight

    def predict_with_gap(
        self,
        imgs: np.ndarray,
        masks: np.ndarray,
        edge_threshold: float = 0.05,
        n_workers: int = 0,
    ):
        """Same function as the original _predict but it calls predict_windows with the model delta_t"""
        # Note: It lags a bit behind the 0.4.0 implem that has added new parameters to _predict

        print("Predicting weights for candidate graph")  # Use print as ByoTrack do not use logging yet
        imgs = normalize(imgs)
        self.transformer.eval()
        features = get_features(
            detections=masks,
            imgs=imgs,
            ndim=self.transformer.config["coord_dim"],
            n_workers=n_workers,
            progbar_class=lambda iterable, **kwargs: tqdm.tqdm(iterable, **kwargs, dynamic_ncols=True),
        )
        print("Building windows")
        kwargs = {"as_torch": True} if version.parse(trackastra.__version__) >= version.parse("0.4.0") else {}
        windows = build_windows(
            features,
            window_size=self.transformer.config["window"],
            progbar_class=lambda iterable, **kwargs: tqdm.tqdm(iterable, **kwargs, dynamic_ncols=True),
            **kwargs,
        )
        print("Predicting windows")
        predictions = predict_windows(
            windows=windows,
            features=features,
            model=self.transformer,
            intra_window_weight=self.intra_weight,
            edge_threshold=edge_threshold,
            spatial_dim=masks.ndim - 1,
            progbar_class=lambda iterable, **kwargs: tqdm.tqdm(iterable, **kwargs, dynamic_ncols=True),
            delta_t=self.delta_t,
        )

        return predictions


@dataclasses.dataclass
class TrackOnStraParameters(FrameByFrameLinkerParameters):
    """Parameters of TrackOnStraLinker

    Attributes:
        association_threshold (float) : Minimum probability to consider a link.
            By default, we use the value from Trackstra.
            Default: 0.05
        positional_cutoff (float): Defines an Euclidean threshold on links.
            We use the default value provided by Trackastra. Tuning it improves performance.
            Default: 256.0
        n_valid (int): Number associated detections required to validate the track after its creation.
            Default: 2
        n_gap (int): Number of consecutive frames without association before the track termination.
            Default: 3
        association_method (AssociationMethod): The frame-by-frame association to use. See `AssociationMethod`.
            It can be provided as a string. (Choice: GREEDY, [SPARSE_]OPT_HARD, [SPARSE_]OPT_SMOOTH)
            Default: OPT_SMOOTH
        anisotropy (Tuple[float, float, float]): Anisotropy of images (Ratio of the pixel sizes
            for each axis, depth first). This will be used to scale distances.
            Default: (1., 1., 1.)
        split_factor (float): Allow splitting of tracks, using a second association step.
            The association threshold in this case is `split_factor * association_threshold`.
            Default: 0.0 (No splits)
        merge_factor (float): Allow merging of tracks, using a second association step.
            The association threshold in this case is `merge_factor * association_threshold`.
            Default: 0.0 (No merges)
    """

    def __init__(
        self,
        *,
        association_threshold: float = 0.05,
        positional_cutoff: float = 256.0,
        n_valid=2,
        n_gap=3,
        association_method: Union[str, AssociationMethod] = AssociationMethod.OPT_SMOOTH,
        anisotropy: Tuple[float, float, float] = (1.0, 1.0, 1.0),
        split_factor: float = 0.0,
        merge_factor: float = 0.0,
    ):
        super().__init__(  # pylint: disable=duplicate-code
            association_threshold=association_threshold,
            n_gap=n_gap,
            n_valid=n_valid,
            association_method=association_method,
            anisotropy=anisotropy,
            split_factor=1 if split_factor > 0 else 0,
            merge_factor=merge_factor,
        )
        self.positional_cutoff = positional_cutoff

    positional_cutoff: float = 256.0


class TrackOnStraLinker(FrameByFrameLinker):
    """Online TrackAstra

    It uses a trained TrackAstra model to predict linking costs. But it replaces the graph optimization from TrackAstra
    by our online FrameByFrame linking. This allows a simple support for false negative detections that TrackAstra
    do not support by itself.

    Warning: This implementation is not yet Online. Indeed the linker has to be setup with
             the full video and detections_sequence before being usable.

    Note:
        This implementation requires trackastra. (pip install trackastra)
        trackastra is only available for python >= 3.10

    See `FrameByFrameLinker` for the other attributes.

    Attributes:
        specs (TrackOnStraParameters): Parameters specifications of the algorithm.
            See `TrackOnStraParameters`.
        model (TrackastraFlex): Model of Trackastra used to compute the association costs.
        cost_dict (Dict[Tuple[int, int, int, int], float]): Cost dictionnary of the feasible tracking graph.

    """

    progress_bar_description = "TrackaOnStra linking"

    def __init__(
        self,
        specs: TrackOnStraParameters,
        model: Optional[TrackastraFlex] = None,
        optflow: Optional[byotrack.OpticalFlow] = None,
        features_extractor: Optional[byotrack.FeaturesExtractor] = None,
    ) -> None:
        super().__init__(specs, optflow, features_extractor)
        self.specs: TrackOnStraParameters
        if model is None:
            self.model = TrackastraFlex.from_pretrained("ctc")
        else:
            self.model = model

        self.model.transformer.config["spatial_pos_cutoff"] = self.specs.positional_cutoff
        self.model.delta_t = self.specs.n_gap + 1
        if self.specs.n_gap == 0:  # If n_gap =0 window size of 4 seems to be a bit better
            self.model.transformer.config["window"] = 4
        elif self.specs.n_gap <= 3:
            self.model.transformer.config["window"] = 5
        else:
            warnings.warn("delta_t is too big, results might not be good")
            self.model.transformer.config["window"] = self.specs.n_gap + 2

        # Initialize empty cost_dict
        self.cost_dict: Dict[Tuple[int, int, int], Dict[int, float]] = {}

        if optflow is not None:
            warnings.warn("OpticalFlow is not supported by this linker, it will be ignored")
        if features_extractor is not None:
            warnings.warn("Features Extractor is not supported by this linker, it will be ignored")

    def setup(self, video: Union[Sequence[np.ndarray], np.ndarray], detections_sequence: Sequence[byotrack.Detections]):
        """Offline setup of the linker by computing all the linking costs with Trackastra.

        Linking costs are stored into `cost_dict`.

        This function needs to be called on each video for the linking to be able to run.

        Args:
            video (Sequence[np.ndarray] | np.ndarray): Sequence of T frames (array).
                Each frame is expected to have a shape ([D, ]H, W, C)
            detections_sequence (Sequence[byotrack.Detections]): Detections for each frame
                Detections is expected for each frame of the video, in the same order.
                (Note that for a given frame, the Detections can be empty)

        """
        # Convert Byotrack format to TrackAstra
        imgs = np.asarray(video)[..., 0]
        masks = np.zeros(imgs.shape, dtype=np.uint16)

        for i, detection in enumerate(detections_sequence):
            masks[i] = detection.segmentation.cpu().numpy().astype(np.uint16)

        # Then compute the cost
        predictions = self.model.predict_with_gap(imgs, masks, edge_threshold=self.specs.association_threshold)

        nodes = predictions["nodes"]
        weights = predictions["weights"]

        # And convert back to a more usable format
        self.cost_dict = build_cost_dict(nodes, weights)

    def motion_model(self) -> None:
        pass  # No motion to model

    def post_association(self, _: np.ndarray, detections: byotrack.Detections, active_mask: torch.Tensor):
        # Simply store the position of the detection
        active_positions = torch.full((len(self.active_tracks), detections.dim), torch.nan)

        for i, track in enumerate(self.active_tracks):
            j = track.detection_ids[-1]
            if j != -1:
                active_positions[i] = detections.position[j]

        self.all_positions.append(active_positions)

    def cost(self, _: np.ndarray, detections: byotrack.Detections) -> Tuple[torch.Tensor, float]:
        """Compute the association cost between active tracks and detections.

        Args:
            detections (byotrack.Detections) : Detections on a given frame.

        Returns:
            torch.Tensor: The cost matrix between active tracks and detections
                Shape: (n_tracks, n_dets), dtype: float
            float: The association threshold to use. Simply returns -ln(association_treshold)
        """
        n = len(self.active_tracks)
        m = detections.length
        cost = torch.full((n, m), torch.inf)
        for index, track in enumerate(self.active_tracks):
            # Find most recent valid detection
            for tau in range(1, self.specs.n_gap + 2):
                if len(track.detection_ids) >= tau and track.detection_ids[-tau] != -1:
                    frame_prev = self.frame_id - tau
                    det_prev = track.detection_ids[-tau]

                    # Direct lookup: all possible detections from cost_dict
                    edges = self.cost_dict.get((frame_prev, det_prev, self.frame_id), {})
                    for index_det, edge_cost in edges.items():
                        cost[index, index_det] = edge_cost
                    break

        return (cost, -np.log(self.specs.association_threshold))



"""Helpers to handle Sinetra dataset [11]

We only provide functions to load annotations.
"""

import os
import pathlib
from typing import List, Dict, Union

import torch

import byotrack


def load_metadata(path: Union[str, os.PathLike]) -> Dict[str, torch.Tensor]:
    """Load all the metadata of the ground-truths

    It loads the outputed metadata from SINETRA (position, size, rotation and intensity weight)
    See: https://github.com/raphaelreme/SINETRA

    Format:

    .. code-block:: json

        {
            "mu": torch.Tensor (T, N, D)  # Position
            "std": torch.Tensor (T, N, D)  # Size
            "theta": torch.Tensor (T, N, 1 or D)  # Rotation (1 in 2D, 3 in 3D)
            "weight": torch.Tensor (T, N)  # Weight
        }

    Args:
        path (str | os.PathLike): Path to the generated folder or to the `video_data.pt` file.

    Returns:
        Dict[str, torch.Tensor]: Metadata (position, size, rotation, and intensity weight)
    """
    path = pathlib.Path(path)

    if path.is_dir():
        return torch.load(path / "video_data.pt", weights_only=True)

    return torch.load(path, weights_only=True)


def load_tracks(path: pathlib.Path) -> List[byotrack.Track]:
    """Load ground-truth tracks, built from the metadata

    This is quite simple, it uses only the positional metadata ("mu") to build tracks.
    Each track is defined from frame 0 to the end in current implementation of SINETRA.

    Args:
        path (str | os.PathLike): Path to the generated folder or to the `video_data.pt` file.

    Returns:
        List[byotrack.Track]: Ground-truth tracks

    """

    ground_truth = load_metadata(path)

    tracks = []
    for i in range(ground_truth["mu"].shape[1]):
        tracks.append(byotrack.Track(0, ground_truth["mu"][:, i], i))

    return tracks
